import typing as t
from urllib.parse import urlparse

import mmh3
from channels.db import database_sync_to_async as db  # type: ignore
from django.db import models
from django.http.request import HttpRequest, QueryDict
from uuid6 import uuid7

from . import json

if t.TYPE_CHECKING:
    T = t.TypeVar("T")
    P = t.ParamSpec("P")

    def db(f: t.Callable[P, T]) -> t.Callable[P, t.Awaitable[T]]: ...  # noqa: UP047


def get_params(obj: HttpRequest | QueryDict | str | None) -> QueryDict:
    if isinstance(obj, HttpRequest):
        is_htmx_request = json.loads(obj.META.get("HTTP_HX_REQUEST", "false"))
        if is_htmx_request:
            return QueryDict(
                urlparse(obj.META["HTTP_HX_CURRENT_URL"]).query,
                mutable=True,
            )
        else:
            qd = QueryDict(None, mutable=True)
            qd.update(obj.GET)
            return qd
    elif isinstance(obj, QueryDict):
        qd = QueryDict(None, mutable=True)
        qd.update(obj)  # type: ignore
        return qd
    elif isinstance(obj, str):
        return QueryDict(
            query_string=urlparse(obj).query if obj else None,
            mutable=True,
        )
    else:
        return QueryDict(None, mutable=True)


def get_instance_subscriptions(
    obj: models.Model,
    actions: t.Sequence[str] = ("created", "updated", "deleted"),
):
    """Get the subscriptions to actions of a single instance of a model.

    This won't return model-level subscriptions.

    The `actions` is the set of actions to subscribe to, including any possible relation (e.g
    'users.deleted').  If actions is empty, return only instance-level subscription.

    """
    cls = type(obj)
    app = cls._meta.app_label
    name = cls._meta.model_name
    prefix = f"{app}.{name}.{obj.pk}"
    if not actions:
        return {prefix}
    else:
        return {f"{prefix}.{action}" for action in actions}


def get_model_subscriptions(
    obj: type[models.Model] | models.Model,
    actions: t.Sequence[str | None] = (),
) -> set[str]:
    """Get the subscriptions to actions of the model.

    If the `obj` is an instance of the model, return all the subscriptions
    from actions.  If `obj` is just the model class, return the top-level
    subscription.

    The `actions` is the set of actions to subscribe to, including any
    possible relation (e.g 'users.deleted').

    """
    actions = actions or (None,)
    if isinstance(obj, models.Model):
        cls = type(obj)
        instance = obj
    else:
        cls = obj
        instance = None
    app = cls._meta.app_label
    name = cls._meta.model_name
    model_prefix = f"{app}.{name}"
    prefix = f"{model_prefix}.{instance.pk}" if instance else model_prefix
    result = {(f"{prefix}.{action}" if action else prefix) for action in actions}
    return result


def generate_id():
    return f"hx-{uuid7().hex}"


def compact_hash(value: str) -> str:
    """Return a SHA1 using a base with 64+ symbols"""
    # this returns a signed 32 bit number, we convert it to unsigned with `& 0xffffffff`
    hashed_value = mmh3.hash(value) & 0xFFFFFFFF

    # Convert the integer to the custom base
    base_len = len(_BASE)
    encoded = []
    while hashed_value > 0:
        hashed_value, rem = divmod(hashed_value, base_len)
        encoded.append(_BASE[rem])

    return "".join(encoded)


# The order of the base is random so that it doesn't match anything out there.
# The symbols are chosen to avoid extra encoding in the URL and HTML, and
# allowed in plain CSS selectors.
_BASE = "ZmBeUHhTgusXNW_Y1b05KPiFcQJD86joqnIRE7Lfkrdp3AOMCvltSwzVG9yxa42"



from django.apps import AppConfig
from django.utils.module_loading import autodiscover_modules


class App(AppConfig):
    name = "djhtmx"
    verbose_name = "Django HTMX"

    def ready(self):
        autodiscover_modules("live")
        autodiscover_modules("htmx")



from django.http import HttpRequest

from djhtmx.repo import Repository


def component_repo(request: HttpRequest):
    return {"htmx_repo": getattr(request, "htmx_repo", None) or Repository.from_request(request)}



from functools import partial
from http import HTTPStatus
from typing import assert_never

from django.apps import apps
from django.core.signing import Signer
from django.http.request import HttpRequest
from django.http.response import HttpResponse
from django.urls import path, re_path
from django.utils.html import format_html

from .commands import PushURL, ReplaceURL, SendHtml
from .component import REGISTRY, Destroy, DispatchDOMEvent, Focus, Open, Redirect, Triggers
from .consumer import Consumer
from .introspection import parse_request_data
from .repo import Repository
from .tracing import tracing_span

signer = Signer()


def endpoint(request: HttpRequest, component_name: str, component_id: str, event_handler: str):
    if "HTTP_HX_SESSION" not in request.META:
        return HttpResponse("Missing header HX-Session", status=HTTPStatus.BAD_REQUEST)

    with tracing_span(f"{component_name}.{event_handler}"):
        repo = Repository.from_request(request)
        content: list[str] = []
        headers: dict[str, str] = {}
        triggers = Triggers()

        for command in repo.dispatch_event(
            component_id,
            event_handler,
            parse_request_data(request.POST | request.FILES)  # type: ignore[reportOperatorIssues]
            | (
                {"prompt": prompt}
                if (prompt := request.META.get("HTTP_HX_PROMPT", None)) is not None
                else {}
            ),
        ):
            # Command loop
            match command:
                case Destroy(component_id):
                    content.append(
                        format_html(
                            '<div hx-swap-oob="outerHtml:#{component_id}"></div>',
                            component_id=component_id,
                        )
                    )
                case Redirect(url):
                    headers["HX-Redirect"] = url
                case Focus(selector):
                    triggers.after_settle("hxFocus", selector)
                case Open(url, name, target, rel):
                    triggers.after_settle(
                        "hxOpenURL",
                        {"url": url, "name": name, "target": target, "rel": rel},
                    )
                case DispatchDOMEvent(target, event, detail, bubbles, cancelable, composed):
                    triggers.after_settle(
                        "hxDispatchDOMEvent",
                        {
                            "event": event,
                            "target": target,
                            "detail": detail,
                            "bubbles": bubbles,
                            "cancelable": cancelable,
                            "composed": composed,
                        },
                    )
                case SendHtml(html):
                    content.append(html)
                case PushURL(url):
                    headers["HX-Push-Url"] = url
                case ReplaceURL(url):
                    headers["HX-Replace-Url"] = url
                case _ as unreachable:
                    assert_never(unreachable)

        return HttpResponse("\n\n".join(content), headers=headers | triggers.headers)


APP_CONFIGS = sorted(apps.app_configs.values(), key=lambda app_config: -len(app_config.name))


def app_name_of_component(cls: type):
    cls_module = cls.__module__
    for app_config in APP_CONFIGS:
        if cls_module.startswith(app_config.name):
            return app_config.label
    return cls_module


urlpatterns = [
    path(
        f"{app_name_of_component(component)}/{component_name}/<component_id>/<event_handler>",
        partial(endpoint, component_name=component_name),
        name=f"djhtmx.{component_name}",
    )
    for component_name, component in REGISTRY.items()
]


ws_urlpatterns = [
    re_path("ws", Consumer.as_asgi(), name="djhtmx.ws"),  # type: ignore
]



import datetime
import enum
import inspect
import operator
import types
from collections import defaultdict
from collections.abc import Callable, Iterable, Sequence
from dataclasses import dataclass
from datetime import date
from functools import cache
from inspect import Parameter, _ParameterKind
from typing import (
    Annotated,
    Any,
    Generic,
    TypedDict,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
    is_typeddict,
)
from uuid import UUID

from django.apps import apps
from django.db import models
from django.db.models import Prefetch
from django.utils.datastructures import MultiValueDict
from pydantic import BeforeValidator, PlainSerializer, TypeAdapter

M = TypeVar("M", bound=models.Model)


@dataclass(slots=True)
class ModelRelatedField:
    name: str
    relation_name: str
    related_model_name: str


MODEL_RELATED_FIELDS: dict[type[models.Model], tuple[ModelRelatedField, ...]] = {}


@dataclass(slots=True, unsafe_hash=True)
class ModelConfig:
    """Annotation to configure fetching the models/querysets in pydantic models.

    For this configuration to take place the pydantic model has to call `annotate_model`.  HTMX
    components require no extra steps.

    """

    lazy: bool = False
    """If set to True, annotations of models.Model will return a _LazyModelProxy instead of the
       actual model instance.

    """

    select_related: list[str] | tuple[str, ...] | None = None
    """The arguments to `model.objects.select_related(*select_related)`."""

    prefetch_related: list[str | Prefetch] | tuple[str | Prefetch, ...] | None = None
    """The arguments to `model.objects.prefetch_related(*prefetch_related)`."""


_DEFAULT_MODEL_CONFIG = ModelConfig()


@dataclass(slots=True, init=False)
class _LazyModelProxy(Generic[M]):  # noqa
    """Deferred proxy for a Django model instance; only fetches from the database on access."""

    __model: type[M]
    __instance: M | None
    __pk: Any | None
    __select_related: Sequence[str] | None
    __prefetch_related: Sequence[str | Prefetch] | None

    def __init__(
        self,
        model: type[M],
        value: Any,
        model_annotation: ModelConfig | None = None,
    ):
        self.__model = model
        if value is None or isinstance(value, model):
            self.__instance = value
            self.__pk = getattr(value, "pk", None)
        else:
            self.__instance = None
            self.__pk = value
        if model_annotation:
            self.__select_related = model_annotation.select_related
            self.__prefetch_related = model_annotation.prefetch_related
        else:
            self.__select_related = None
            self.__prefetch_related = None

    def __getattr__(self, name: str) -> Any:
        if name == "pk":
            return self.__pk
        if self.__instance is None:
            self.__ensure_instance()
        return getattr(self.__instance, name)

    def __ensure_instance(self):
        if not self.__instance:
            manager = self.__model.objects
            if select_related := self.__select_related:
                manager = manager.select_related(*select_related)
            if prefetch_related := self.__prefetch_related:
                manager = manager.prefetch_related(*prefetch_related)
            self.__instance = manager.get(pk=self.__pk)
        return self.__instance

    def __repr__(self) -> str:
        return f"<_LazyModelProxy model={self.__model}, pk={self.__pk}, instance={self.__instance}>"


@dataclass(slots=True)
class _ModelBeforeValidator(Generic[M]):  # noqa
    model: type[M]
    model_config: ModelConfig

    def __call__(self, value):
        if self.model_config.lazy:
            return self._get_lazy_proxy(value)
        else:
            return self._get_instance(value)

    def _get_lazy_proxy(self, value):
        if isinstance(value, _LazyModelProxy):
            instance = value._LazyModelProxy__instance or value._LazyModelProxy__pk
            return _LazyModelProxy(self.model, instance)
        else:
            return _LazyModelProxy(self.model, value)

    def _get_instance(self, value):
        if value is None or isinstance(value, self.model):
            return value
        # If a component has a lazy model proxy, and passes it down to another component that
        # doesn't allow lazy proxies, we need to materialize it.
        elif isinstance(value, _LazyModelProxy):
            return value._LazyModelProxy__ensure_instance()
        else:
            manager = self.model.objects
            if select_related := self.model_config.select_related:
                manager = manager.select_related(*select_related)
            if prefetch_related := self.model_config.prefetch_related:
                manager = manager.prefetch_related(*prefetch_related)
            return manager.get(pk=value)

    @classmethod
    @cache
    def from_modelclass(cls, model: type[M], model_config: ModelConfig):
        return cls(model, model_config=model_config)


@dataclass(slots=True)
class _ModelPlainSerializer(Generic[M]):  # noqa
    model: type[M]

    def __call__(self, value):
        return value.pk

    @classmethod
    @cache
    def from_modelclass(cls, model: type[M]):
        return cls(model)


def _Model(model: type[models.Model], model_config: ModelConfig | None = None):
    assert issubclass_safe(model, models.Model)
    model_config = model_config or _DEFAULT_MODEL_CONFIG
    return Annotated[
        model if not model_config.lazy else _LazyModelProxy[model],
        BeforeValidator(_ModelBeforeValidator.from_modelclass(model, model_config)),
        PlainSerializer(
            func=_ModelPlainSerializer.from_modelclass(model),
            return_type=guess_pk_type(model),
        ),
    ]


def _QuerySet(qs: type[models.QuerySet]):
    [model] = [m for m in apps.get_models() if isinstance(m.objects.all(), qs)]
    return Annotated[
        qs,
        BeforeValidator(lambda v: (v if isinstance(v, qs) else model.objects.filter(pk__in=v))),
        PlainSerializer(
            func=lambda v: (
                [instance.pk for instance in v]
                if v._result_cache
                else list(v.values_list("pk", flat=True))
            ),
            return_type=guess_pk_type(model),
        ),
    ]


def annotate_model(annotation, *, model_config: ModelConfig | None = None):
    if issubclass_safe(annotation, models.Model):
        return _Model(annotation, model_config)
    elif issubclass_safe(annotation, models.QuerySet):
        return _QuerySet(annotation)
    elif is_typeddict(annotation):
        return TypedDict(
            annotation.__name__,  # type: ignore
            {
                k: annotate_model(v)  # type: ignore
                for k, v in get_type_hints(annotation).items()
            },
        )
    elif type_ := get_origin(annotation):
        if type_ is types.UnionType or type_ is Union:
            type_ = Union
        match get_args(annotation):
            case ():
                return type_
            case (param,):
                return type_[annotate_model(param)]  # type: ignore
            case params:
                model_annotation = next(
                    (p for p in params if isinstance(p, ModelConfig)),
                    None,
                )
                return type_[*(annotate_model(p, model_config=model_annotation) for p in params)]  # type: ignore
    else:
        return annotation


def guess_pk_type(model: type[models.Model]):
    match model._meta.pk:
        case models.UUIDField():
            return UUID
        case models.IntegerField():
            return int
        case _:
            return str


def isinstance_safe(o, types):
    try:
        return isinstance(o, types)
    except TypeError:
        return False


def issubclass_safe(o, types):
    try:
        return issubclass(o, types)
    except TypeError:
        return False


# for state of old components


def get_function_parameters(
    function: Callable,
    exclude_kinds: tuple[_ParameterKind, ...] = (),
) -> frozenset[str]:
    return frozenset(
        param.name
        for param in inspect.signature(function).parameters.values()
        if param.name != "self" and param.kind not in exclude_kinds
    )


@cache
def get_related_fields(model):
    related_fields = MODEL_RELATED_FIELDS.get(model)
    if related_fields is None:
        fields = []
        for field in model._meta.get_fields():
            if (
                isinstance(field, models.ForeignKey)
                and (relation_name := field.related_query_name())
                and relation_name != "+"
            ):
                rel_meta = field.related_model._meta  # type: ignore
                fields.append(
                    ModelRelatedField(
                        name=field.attname,
                        relation_name=relation_name,
                        related_model_name=(f"{rel_meta.app_label}.{rel_meta.model_name}"),
                    )
                )
        related_fields = MODEL_RELATED_FIELDS[model] = tuple(fields)
    return related_fields


# filtering


def filter_parameters(f: Callable, kwargs: dict[str, Any]):
    has_kwargs = any(
        param.kind == Parameter.VAR_KEYWORD for param in inspect.signature(f).parameters.values()
    )
    if has_kwargs:
        return kwargs
    else:
        return {
            param: value
            for param, value in kwargs.items()
            if param in inspect.signature(f).parameters
        }


# Decoder for client requests


def parse_request_data(data: MultiValueDict[str, Any] | dict[str, Any]):
    if not isinstance(data, MultiValueDict):
        data = MultiValueDict({
            key: value if isinstance(value, list) else [value] for key, value in data.items()
        })
    return _parse_obj(_extract_data(data))


def _extract_data(data: MultiValueDict[str, Any]):
    for key in set(data):
        if key.endswith("[]"):
            value = data.getlist(key)
            key = key.removesuffix("[]")
        else:
            value = data.get(key)
        yield key.split("."), value


def _parse_obj(data: Iterable[tuple[list[str], Any]], output=None) -> dict[str, Any] | Any:
    output = output or {}
    arrays = defaultdict(lambda: defaultdict(dict))  # field -> index -> value
    for key, value in data:
        fragment, *tail = key
        if "[" in fragment:
            field_name = fragment[: fragment.index("[")]
            index = int(fragment[fragment.index("[") + 1 : -1])
            arrays[field_name][index] = (
                _parse_obj([(tail, value)], arrays[field_name][index]) if tail else value
            )
        else:
            output[fragment] = _parse_obj([(tail, value)]) if tail else value

    for field, items in arrays.items():
        output[field] = [v for _, v in sorted(items.items(), key=operator.itemgetter(0))]
    return output


def get_event_handler_event_types(f: Callable[..., Any]) -> set[type]:
    "Extract the types of the annotations of parameter 'event'."
    event = get_type_hints(f)["event"]
    origin = get_origin(event)
    if origin is types.UnionType or origin is Union:
        return {
            arg for arg in get_args(event) if isinstance(arg, type) and arg is not types.NoneType
        }
    elif isinstance(event, type):
        return {event}
    else:
        return set()


def get_annotation_adapter(annotation):
    """Return a TypeAdapter for the annotation."""
    if annotation is bool:
        return infallible_bool_adapter

    return TypeAdapter(annotation, config={"arbitrary_types_allowed": True})


# Infallible adapter for boolean values.  't' is True, everything else is
# False.
infallible_bool_adapter = TypeAdapter(
    Annotated[
        bool,
        BeforeValidator(lambda v: v == "t"),
        PlainSerializer(lambda v: "t" if v else "f"),
    ]
)


def is_basic_type(ann):
    """Returns True if the annotation is a simple type.

    Simple types are:

    - Simple Python types: ints, floats, strings, UUIDs, dates and datetimes, bools,
      and the value None.

    - Instances of a Django model (which will use the PK as a proxy)

    - Instances of IntEnum or StrEnum.

    - Instances of dict, tuple, list or set

    """
    return (
        ann in _SIMPLE_TYPES
        #  __origin__ -> model in 'Annotated[model, BeforeValidator(...), PlainSerializer(...)]'
        or issubclass_safe(getattr(ann, "__origin__", None), models.Model)
        or issubclass_safe(ann, (enum.IntEnum, enum.StrEnum))
        or is_collection_annotation(ann)
    )


def is_union_of_basic(ann):
    """Returns True Union of simple types (as is_simple_annotation)"""
    type_ = get_origin(ann)
    if type_ is types.UnionType or type_ is Union:
        return all(is_basic_type(arg) for arg in get_args(ann))
    return False


def is_simple_annotation(ann):
    "Return True if the annotation is either simple or a Union of simple"
    return is_basic_type(ann) or is_union_of_basic(ann)


def is_collection_annotation(ann):
    return issubclass_safe(ann, _COLLECTION_TYPES)


Unset = object()
_SIMPLE_TYPES = (int, str, float, UUID, types.NoneType, date, datetime, bool)
_COLLECTION_TYPES = (dict, tuple, list, set)



import sys
from collections import defaultdict

import djclick as click
from xotl.tools.future.itertools import delete_duplicates
from xotl.tools.objects import get_branch_subclasses as get_final_subclasses

from djhtmx.component import REGISTRY, HtmxComponent


def bold(msg):
    return click.style(str(msg), bold=True)


def yellow(msg):
    return click.style(str(msg), fg="yellow")


@click.group()
def htmx():
    pass


@htmx.command("check-missing")  # type: ignore
@click.argument("fname", type=click.File())
def check_missing(fname):
    r"""Check if there are any missing HTMX components.

    Expected usage:

    find -type f -name '*.html' | while read f; do grep -P '{% htmx .(\w+)' -o $f \
    | awk '{print $3}' | cut -b2-; done | sort -u \
    | python manage.py htmx check-missing -

    """
    names = {n.strip() for n in fname.readlines()}
    known = set(REGISTRY)
    missing = list(names - known)
    if missing:
        missing.sort()
        for n in missing:
            click.echo(
                f"Missing component detected {bold(yellow(n))}",
                file=sys.stderr,
            )
        sys.exit(1)


@htmx.command("check-shadowing")  # type: ignore
def check_shadowing():
    "Checks if there are components that might shadow one another."
    clashes = defaultdict(list)
    for cls in get_final_subclasses(
        HtmxComponent,  # type: ignore
        without_duplicates=True,
    ):
        name = cls.__name__
        registered = REGISTRY.get(name)
        if registered is not cls and registered is not None:
            clashes[name].append(cls)
            clashes[name].append(registered)

    if clashes:
        for name, shadows in clashes.items():
            shadows = delete_duplicates(shadows)
            if shadows:
                click.echo(f"HtmxComponent {bold(name)} might be shadowed by:")
                for shadow in shadows:
                    click.echo(f"  -  {bold(shadow.__module__)}.{bold(shadow.__name__)}")

        sys.exit(1)



import asyncio
from collections.abc import Awaitable, Callable

from asgiref.sync import sync_to_async
from django.http import HttpRequest, HttpResponse


def middleware(
    get_response: Callable[[HttpRequest], HttpResponse]
    | Callable[[HttpRequest], Awaitable[HttpResponse]],
):
    """
    Middleware function that wraps get_response and ensures the HTMX repository
    is flushed and removed from the request after handling each request. It can handle
    both sync and async get_response automatically.
    """

    if asyncio.iscoroutinefunction(get_response):
        # Async version
        async def middleware(request: HttpRequest) -> HttpResponse:  # type: ignore
            response = await get_response(request)
            if repo := getattr(request, "htmx_repo", None):
                await sync_to_async(repo.session.flush)()
                delattr(request, "htmx_repo")
            return response

    else:
        # Sync version
        def middleware(request: HttpRequest) -> HttpResponse:
            response = get_response(request)
            if repo := getattr(request, "htmx_repo", None):
                repo.session.flush()
                delattr(request, "htmx_repo")
            return response  # type: ignore

    return middleware



from __future__ import annotations

import logging
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Literal

from django.db import models
from django.http import QueryDict
from django.shortcuts import resolve_url
from django.utils.safestring import SafeString

logger = logging.getLogger(__name__)


@dataclass(slots=True)
class SendHtml:
    content: SafeString

    # debug trace for troubleshooting
    debug_trace: str | None = None


@dataclass(slots=True)
class PushURL:
    url: str
    command: Literal["push_url"] = "push_url"

    @classmethod
    def from_params(cls, params: QueryDict):
        return cls("?" + params.urlencode())

    @classmethod
    def to(cls, to: Callable[..., Any] | models.Model | str, *args, **kwargs):
        return cls(resolve_url(to, *args, **kwargs))


@dataclass(slots=True)
class ReplaceURL:
    url: str
    command: Literal["replace_url"] = "replace_url"

    @classmethod
    def from_params(cls, params: QueryDict):
        return cls("?" + params.urlencode())

    @classmethod
    def to(cls, to: Callable[..., Any] | models.Model | str, *args, **kwargs):
        return cls(resolve_url(to, *args, **kwargs))



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Any

from django.http import QueryDict
from pydantic import BaseModel, TypeAdapter
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined

from djhtmx.introspection import (
    get_annotation_adapter,
    is_collection_annotation,
    is_simple_annotation,
)
from djhtmx.utils import compact_hash


@dataclass(slots=True, unsafe_hash=True)
class Query:
    """Annotation to integrate the state with the URL's query string.

    By default the query string name can be shared across many components,
    provided the have the same type annotation.

    You can set `shared` to False, to make this a specific (by component id)
    param.  In this case the URL is `<name>__<ns>=value`.

    If `auto_subscribe` is True (the default), the component is automatically
    subscribed to changes in the query string.  Otherwise, changes in the
    query string won't be signaled.

    """

    name: str
    shared: bool = True
    auto_subscribe: bool = True

    def __post_init__(self):
        assert _VALID_QS_NAME_RX.match(self.name) is not None, self.name

    @classmethod
    def extract_from_field_info(cls, name: str, field: FieldInfo):
        done = False
        for meta in field.metadata:
            if isinstance(meta, cls):
                if done:
                    raise TypeError(
                        f"Field '{name}' in component {cls.__qualname__} "
                        " has more than one Query annotation."
                    )
                if not (
                    field.default is not PydanticUndefined or field.default_factory is not None
                ):
                    raise TypeError(
                        f"Field '{name}' of {cls.__qualname__} must have "
                        "a default or default_factory."
                    )

                yield meta
                done = True


@dataclass(slots=True)
class QueryPatcher:
    field_name: str
    param_name: str
    signal_name: str
    auto_subscribe: bool

    default_value: Any
    adapter: TypeAdapter[Any]

    use_json: bool

    @classmethod
    def for_component(cls, component: type[BaseModel]):
        seen = set()
        for field_name, field in component.model_fields.items():
            for query in Query.extract_from_field_info(field_name, field):
                name = query.name
                if name in seen:
                    raise TypeError(
                        f"Component {component.__name__} has multiple "
                        f"fields with the same query param '{name}'"
                    )
                seen.add(name)

                # Check the type annotation.  It must be something that can
                # reasonably be put in the URL: basic types or union of basic
                # types.
                annotation = field.annotation
                if not is_simple_annotation(annotation):
                    raise TypeError(f"Invalid type annotation {annotation} for a query string")

                # The field must have a default to be Query.
                if field.default is PydanticUndefined and field.default_factory is None:
                    raise TypeError(
                        f"Field '{name}' of {component.__name__} must have "
                        "a default or default_factory."
                    )

                # Convert parameter from `search_query` to `search-query`
                param_name = name.replace("_", "-")

                # Prefix with the component name if not shared
                if not query.shared:
                    param_name = f"{param_name}-{compact_hash(component.__name__)}"
                adapter = get_annotation_adapter(field.annotation)
                yield cls(
                    field_name=field_name,
                    param_name=param_name,
                    signal_name=f"querystring.{param_name}",
                    auto_subscribe=query.shared and query.auto_subscribe,
                    default_value=field.get_default(call_default_factory=True),
                    adapter=adapter,
                    use_json=is_collection_annotation(annotation),
                )

    def get_update_for_state(self, params: QueryDict):
        if (raw_param := params.get(self.param_name)) is not None:
            # We need to perform the validation during patching, otherwise
            # ill-formed values in the query will cause a Pydantic
            # ValidationError, but we should just simply ignore invalid
            # values.
            try:
                return {
                    self.field_name: self.adapter.validate_json(raw_param)
                    if self.use_json
                    else self.adapter.validate_python(raw_param)
                }
            except ValueError:
                # Preserve the last good known state in the component
                return {}
        else:
            return {self.field_name: self.default_value}

    def get_updates_for_params(self, value: Any, params: QueryDict) -> list[str]:
        # If we're setting the default value, let remove it from the query
        # string completely, and trigger the signal if needed.
        if value == self.default_value:
            if self.param_name in params:
                params.pop(self.param_name, None)
                return [self.signal_name]
            else:
                return []

        # Otherwise, let's serialize the value and only update it if it is
        # different.
        if self.use_json:
            serialized_value = self.adapter.dump_json(value)
        else:
            serialized_value = self.adapter.dump_python(value, mode="json")
        try:
            # We need to validate and dump back to get the exact JSON-friendly
            # type representation.  Otherwise dates, enums, and other types
            # won't match the serialized value.
            param = params.get(self.param_name)
            if self.use_json:
                previous_value = self.adapter.dump_json(self.adapter.validate_json(param or ""))
            else:
                previous_value = self.adapter.dump_python(
                    self.adapter.validate_python(param),
                    mode="json",
                )
        except ValueError:
            previous_value = self.default_value

        if serialized_value == previous_value:
            return []
        else:
            params[self.param_name] = serialized_value  # type: ignore
            return [self.signal_name]


_VALID_QS_NAME_RX = re.compile(r"^[a-zA-Z_\d][-a-zA-Z_\d]*$")



import dataclasses
import enum
import json
from collections.abc import Generator

import orjson
from django.core.serializers import deserialize, serialize
from django.core.serializers.base import DeserializedObject
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from pydantic import BaseModel

loads = orjson.loads


def dumps(obj):
    return orjson.dumps(obj, default).decode()


def encode(instance: models.Model) -> str:
    return serialize("json", [instance], cls=HtmxEncoder)


def decode(instance: str) -> models.Model:
    obj: DeserializedObject = next(iter(deserialize("json", instance)))
    obj.object.save = obj.save  # type: ignore
    return obj.object


class HtmxEncoder(json.JSONEncoder):
    def default(self, o):
        return default(o)


def default(o):
    try:
        return DjangoJSONEncoder().default(o)
    except TypeError:
        if hasattr(o, "__json__"):
            return o.__json__()

        if isinstance(o, models.Model):
            return o.pk

        if isinstance(o, Generator | set | frozenset):
            return list(o)

        if BaseModel and isinstance(o, BaseModel):
            return o.model_dump()

        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)  # type: ignore

        if isinstance(o, enum.Enum):
            return o.value
        raise



import contextlib

from djhtmx import settings

# pragma: no cover

try:
    import sentry_sdk  # pyright: ignore[reportMissingImports]
except ImportError:
    sentry_sdk = None


if settings.ENABLE_SENTRY_TRACING and sentry_sdk is not None:
    sentry_start_span = sentry_sdk.start_span

    @contextlib.contextmanager
    def _sentry_span(name: str, **tags: str):  # pyright: ignore[reportRedeclaration]
        with sentry_start_span(op="djhtmx", name=name) as span:
            for tag, value in tags.items():
                span.set_tag(tag, value)
            yield span

else:

    @contextlib.contextmanager
    def _sentry_span(description, **tags):
        yield


try:
    import logfire  # pyright: ignore[reportMissingImports]
except ImportError:
    logfire = None


if settings.ENABLE_LOGFIRE_TRACING and logfire is not None:
    logfire_span = logfire.span

    def _logfire_span(name: str, **tags):  # pyright: ignore[reportRedeclaration]
        return logfire_span(name, op="djhtmx", **tags)

else:

    @contextlib.contextmanager
    def _logfire_span(description, **tags):
        yield


@contextlib.contextmanager
def tracing_span(name: str, **tags: str):
    with _sentry_span(name, **tags), _logfire_span(name, **tags):
        yield



import random
from collections.abc import Mapping
from typing import Any, Literal, cast

from django import template
from django.core.signing import Signer
from django.template.base import FilterExpression, Node, Parser, Token
from django.template.context import Context
from django.template.defaulttags import TemplateIfParser, TemplateLiteral
from django.template.exceptions import TemplateSyntaxError
from django.urls import reverse
from django.utils.html import format_html, format_html_join
from django.utils.safestring import mark_safe

from .. import json, settings
from ..component import HtmxComponent
from ..repo import Repository

register = template.Library()
signer = Signer()

unset = object()


@register.inclusion_tag(
    "htmx/headers.html",
    takes_context=True,
    name="htmx-headers",
)
def htmx_headers(context):
    """Loads all the necessary scripts to make this work

    Use this tag inside your `<header></header>`.
    """
    if context.get("request"):
        return {
            "enabled": True,
            "SCRIPT_URLS": settings.SCRIPT_URLS,
            "CSRF_HEADER_NAME": settings.CSRF_HEADER_NAME,
            "csrf_token": context.get("csrf_token", ""),
        }
    else:
        return {"enabled": False}


@register.filter(name="add_delay_jitter")
def add_delay_jitter(event, arg=None):
    """Add a random `delay:{jitter}ms` to an event.

    The optional argument is the bound of the jitter.  By default is "100,
    1000" (ms).  The format is always a string with one or two numbers.  If
    the format is not right, fallback to the default.

    Example usage:

    ```
      <div {% on 'load'|add_htmx_jitter:'2000, 30000' 'render' %} ></div>
    ```

    """
    if not arg:
        arg = "100, 1000"
    try:
        min_arg, max_arg = arg.split(",", 1)
        min_ = int(min_arg.strip())
        max_ = int(max_arg.strip())
    except ValueError:
        min_, max_ = 100, 1000
    jitter = random.randint(min_, max_)
    return format_html(f"{event} {{}}", f"delay:{jitter}ms")


@register.simple_tag(takes_context=True)
def htmx(
    context,
    _name: str,
    _state: dict[str, Any] | None = None,
    *,
    lazy: Literal["once"] | bool = False,
    **state,
):
    """Inserts an HTMX Component.

    Pass the component name and the initial state:

        ```html
        {% htmx 'AmazinData' data=some_data %}
        ```
    """
    state = (_state or {}) | state
    repo: Repository = context["htmx_repo"]
    state |= {"lazy": lazy is True}

    # Extract parent component ID from context if available
    parent_id = getattr(context.get("this"), "id", None)

    component = repo.build(_name, state, parent_id=parent_id)
    return repo.render_html(
        component,
        lazy=lazy if isinstance(lazy, bool) else False,
    )


@register.simple_tag(takes_context=True, name="hx-tag")
def hx_tag(context: Context):
    """Adds initialziation data to your root component tag.

    When your component starts, put it there:

        ```html
        {% load htmx %}
        <div {% hx-tag %}>
          ...
        </div>
        ```
    """
    component: HtmxComponent = context["this"]
    oob = context.get("hx_oob")
    context["hx_oob"] = False
    attrs = {
        "id": component.id,
        "hx-swap-oob": "true" if oob else None,
        "hx-headers": json.dumps({"HX-Session": context["htmx_repo"].session_signed_id}),
    }
    if context.get("hx_lazy"):
        context["hx_lazy"] = False
        jitter = random.randint(100, 1000)
        attrs |= {
            "hx-trigger": f"revealed delay:{jitter}ms",
            "hx-get": event_url(component, "render"),
        }
    if settings.DEBUG:
        attrs |= {"hx-name": component.hx_name}
    return format_html_attrs(attrs)


@register.simple_tag(takes_context=True)
def oob(context: Context, suffix: str):
    oob = context.get("hx_oob")
    context["hx_oob"] = False
    id = "-".join(filter(None, (context.get("id"), suffix)))
    return format_html_attrs({"id": id, "hx-swap-oob": "true" if oob else None})


@register.simple_tag(takes_context=True)
def on(
    context,
    _trigger,
    _event_handler=None,
    hx_include: str | None = None,
    **kwargs,
):
    """Binds an event to a handler

    If no trigger is provided, it assumes the default one by omission, in this
    case `click`, for an input is `change`:

        ```html
        <button {% on 'inc' %}>+</button>
        ```

    You can pass it explicitly:

        ```html
        <button {% on 'click' 'inc' %}>+</button>
        ```

    You can also pass explicit arguments:

        ```html
        <button {% on 'click' 'inc' amount=2 %}>+2</button>
        ```

    Remember that together with the explicit arguments, all fields with a
    `name` are passed as implicit arguments to your event handler.

    If you wanna do more advanced stuff read:
    [hx-trigger](https://htmx.org/attributes/hx-trigger/)

    """
    if not _event_handler:
        _event_handler = _trigger
        _trigger = None

    component: HtmxComponent = context["this"]

    if settings.DEBUG:
        assert _event_handler in component._event_handler_params, (
            f"{type(component).__name__}.{_event_handler} event handler not found"
        )

    if not hx_include:
        has_implicit_params = bool(component._event_handler_params[_event_handler] - set(kwargs))
        if has_implicit_params:
            hx_include = f"#{component.id} [name]"

    attrs = {
        "hx-swap": "none",
        "hx-post": event_url(component, _event_handler),
        "hx-trigger": _trigger,
        "hx-vals": json.dumps(kwargs) if kwargs else None,
        "hx-include": hx_include,
    }
    return format_html_attrs(attrs)


def format_html_attrs(attrs: dict[str, Any]):
    return format_html_join(
        "\n",
        '{}="{}"',
        [(k, v) for k, v in attrs.items() if v is not None],
    )


def event_url(component: HtmxComponent, event_handler: str):
    return reverse(
        f"djhtmx.{type(component).__name__}",
        kwargs={
            "component_id": component.id,
            "event_handler": event_handler,
        },
    )


# Shortcuts and helpers

_json_script_escapes = {
    ord(">"): "\\u003E",
    ord("<"): "\\u003C",
    ord("&"): "\\u0026",
}


@register.filter(name="safe_json")
def safe_json(obj):
    return mark_safe(json.dumps(obj).translate(_json_script_escapes))


@register.tag(name="class")
def class_cond(parser: Parser, token: Token):
    """Prints classes conditionally

    ```html
    <div {% class 'btn': True, 'loading': loading, 'falsy': 0} %}></div>
    <div {% class 'btn' x == 'something', 'loading': y is None %}></div>
    ```

    If `loading` is `True` will print:

    ```html
    <div class="btn loading"></div>
    ```
    """
    bits = token.split_contents()[1:]
    classes: list[tuple[FilterExpression, list[str]]] = []

    is_class_name = True
    for bit in bits:
        if is_class_name:
            if bit.endswith(":"):
                classes.append((FilterExpression(bit[:-1], parser), []))
                is_class_name = False
                continue
            else:
                raise TemplateSyntaxError(f"Expected colon (:) after: {bit}")
        else:
            if bit.endswith(","):
                expr = bit[:-1]
                is_class_name = True
            else:
                expr = bit
        classes[-1][1].append(expr)

    return ClassNode([
        (TemplateIfParser(parser, expr).parse(), class_name) for class_name, expr in classes
    ])


class ClassNode(Node):
    def __init__(self, condition_and_classes: list[tuple[TemplateLiteral, FilterExpression]]):
        self.condition_and_classes = condition_and_classes

    def render(self, context: Context):
        class_names = [
            class_name.resolve(cast(Mapping[str, Any], context))
            for condition, class_name in self.condition_and_classes
            if condition.eval(context)  # type: ignore
        ]
        return format_html_attrs({"class": " ".join(class_names) or None})



from dataclasses import dataclass


@dataclass
class HtmxUnhandledError:
    """HTMX triggers this event for any HTMX handler that fails unhandled.

    Applications could subscribe to this event to have last-resource general error recovery
    mechanism.

    """

    error: BaseException | None
    handler_annotations: dict | None = None



from collections import defaultdict
from collections.abc import Callable, Iterable
from functools import reduce
from typing import Any, ParamSpec, TypeVar
from urllib.parse import urlparse

from django.contrib.auth.models import AnonymousUser
from django.test import Client
from lxml import html
from pygments import highlight
from pygments.formatters import TerminalTrueColorFormatter
from pygments.lexers import HtmlLexer

from . import json
from .commands import PushURL, ReplaceURL, SendHtml
from .component import Destroy, DispatchDOMEvent, Focus, HtmxComponent, Open, Redirect
from .introspection import parse_request_data
from .repo import Repository, Session, signer
from .utils import get_params

P = ParamSpec("P")
TPComponent = TypeVar("TPComponent", bound=HtmxComponent)


class Htmx:
    def __init__(self, client: Client):
        self.client = client

    def navigate_to(self, url: str, *args, **kwargs):
        kwargs.setdefault("follow", True)
        response = self.client.get(url, *args, **kwargs)
        assert 200 <= response.status_code < 300
        self.path = response.request["PATH_INFO"]
        self.query_string = response.request["QUERY_STRING"]

        self.dom = html.fromstring(response.content)
        session_id = reduce(
            lambda session, element: (
                session or json.loads(element.attrib["hx-headers"]).get("HX-Session")
            ),
            self.dom.cssselect("[hx-headers]"),
            None,
        )
        assert session_id, "Can't find djhtmx session id"
        session_id = signer.unsign(session_id)

        self.repo = Repository(
            user=response.context.get("user") or AnonymousUser(),
            session=Session(session_id),
            params=get_params(self.query_string),
        )

    def get_component_by_type(self, component_type: type[TPComponent]) -> TPComponent:
        [component] = self.repo.get_components_by_names(component_type.__name__)
        return component  # type: ignore

    def get_components_by_type(self, component_type: type[TPComponent]) -> Iterable[TPComponent]:
        return self.repo.get_components_by_names(component_type.__name__)  # type: ignore

    def get_component_by_id(self, component_id: str):
        component = self.repo.get_component_by_id(component_id)
        assert isinstance(component, HtmxComponent)
        return component

    def type(self, selector: str | html.HtmlElement, text: str, clear=False):
        """Sets the value of an input, by "typing" in to it"""
        element = self._select(selector)
        if (
            element.tag == "input" and element.attrib.get("type", "text") == "text"
        ) or element.tag == "textarea":
            if clear:
                element.attrib["value"] = text
            else:
                element.attrib["value"] = element.attrib.get("value", "") + text
        else:
            assert False, f"Can't type in element {element}"

    def find_by_text(self, text: str) -> html.HtmlElement:
        return self.dom.xpath(f"//*[text()='{text}']")

    def select(self, selector: str) -> list[html.HtmlElement]:
        return self.dom.cssselect(selector)

    def print(self, element: html.HtmlElement):
        print(
            highlight(
                html.tostring(element, pretty_print=True, encoding="unicode"),
                HtmlLexer(),
                TerminalTrueColorFormatter(),
            )
        )

    def _select(self, selector: str | html.HtmlElement) -> html.HtmlElement:
        if isinstance(selector, str):
            [element] = self.dom.cssselect(selector)
        else:
            element = selector
        return element

    def trigger(self, selector: str | html.HtmlElement):
        element = self._select(selector)

        # mutate in case of a checkbox and radios
        match element.tag, element.attrib.get("type"):
            case "input", "checkbox":
                if "checked" in element.attrib:
                    element.attrib.pop("checked")
                else:
                    element.attrib["checked"] = ""
            case "input", "radio":
                if name := element.attrib.get("name"):
                    for radio in self.dom.cssselect(f'input[type=radio][name="{name}"]'):
                        radio.attrib.pop("checked", None)
                element.attrib["checked"] = ""
            case _:
                pass

        [_, component_id, event_handler] = element.attrib["hx-post"].rsplit("/", 2)

        # gather values
        vals = defaultdict(list)
        if include := element.attrib.get("hx-include"):
            for element in self.dom.cssselect(include):
                name = element.attrib["name"]
                value = element.attrib.get("value", "")
                match element.tag, element.attrib.get("type"):
                    case _, "checkbox":
                        if "checked" in element.attrib:
                            vals[name].append(value or "on")
                    case _, "radio":
                        if "checked" in element.attrib:
                            vals[name].append(value)
                    case "select", _:
                        for option in element.cssselect("option[selected]"):
                            vals[name].append(option.attrib.get("value", ""))
                    case _, _:
                        vals[name].append(value)

        vals |= json.loads(element.attrib.get("hx-vals", "{}"))
        self.dispatch_event(component_id, event_handler, parse_request_data(vals))

    def send(self, method: Callable[P, Any], *args: P.args, **kwargs: P.kwargs):
        assert not args, "All parameters have to be passed by name"
        self.dispatch_event(method.__self__.id, method.__name__, kwargs)  # type: ignore

    def dispatch_event(self, component_id: str, event_handler: str, kwargs: dict[str, Any]):
        commands = self.repo.dispatch_event(component_id, event_handler, kwargs)
        navigate_to_url = None
        for command in commands:
            match command:
                case SendHtml(content):
                    incoming = html.fromstring(content)
                    oob: str = incoming.attrib["hx-swap-oob"]
                    if oob == "true":
                        target = self.dom.get_element_by_id(incoming.attrib["id"])
                        if parent := target.getparent():
                            parent.replace(target, incoming)
                    elif oob.startswith("beforeend: "):
                        target_selector = oob.removeprefix("beforeend: ")
                        [target] = self.dom.cssselect(target_selector)
                        target.append(incoming.getchildren()[0])
                    elif oob.startswith("afterbegin: "):
                        target_selector = oob.removeprefix("afterbegin: ")
                        [target] = self.dom.cssselect(target_selector)
                        target.insert(0, incoming.getchildren()[0])
                    elif oob.startswith("afterend: "):
                        target_selector = oob.removeprefix("afterend: ")
                        [target] = self.dom.cssselect(target_selector)
                        target.addnext(incoming.getchildren()[0])
                    elif oob.startswith("beforebegin: "):
                        target_selector = oob.removeprefix("afterend: ")
                        [target] = self.dom.cssselect(target_selector)
                        target.addprevious(incoming.getchildren()[0])
                    else:
                        assert False, "Unknown swap strategy, please define it here"

                case Destroy(component_id):
                    target = self.dom.get_element_by_id(component_id)
                    if parent := target.getparent():
                        parent.remove(target)

                case Redirect(url) | Open(url):
                    navigate_to_url = url

                case PushURL(url) | ReplaceURL(url):
                    parsed_url = urlparse(url)
                    self.path = parsed_url.path
                    self.query_string = parsed_url.query

                case Focus() | DispatchDOMEvent():
                    pass

        if navigate_to_url:
            self.navigate_to(navigate_to_url)



from typing import assert_never

from django.db import models
from django.db.models.signals import post_save, pre_delete

from djhtmx.commands import PushURL, ReplaceURL

from .component import (
    BuildAndRender,
    Command,
    Destroy,
    DispatchDOMEvent,
    Emit,
    Execute,
    Focus,
    Open,
    Redirect,
    Render,
    Signal,
    SkipRender,
)
from .introspection import get_related_fields
from .utils import get_model_subscriptions


class CommandQueue:
    def __init__(self, commands: list[Command]):
        self.processing_component_id: str = ""
        self._commands = commands
        self._destroyed_ids: set[str] = set()
        self._optimize()

        # subscribe to signals changes
        post_save.connect(self._listen_to_post_save_and_pre_delete, weak=True)
        pre_delete.connect(self._listen_to_post_save_and_pre_delete, weak=True)

    def _listen_to_post_save_and_pre_delete(
        self,
        sender: type[models.Model],
        instance: models.Model,
        created: bool | None = None,
        **kwargs,
    ):
        if created is None:
            action = "deleted"
        elif created:
            action = "created"
        else:
            action = "updated"

        signals = get_model_subscriptions(
            instance, actions=(action, None)
        ) | get_model_subscriptions(type(instance), actions=(action, None))

        for field in get_related_fields(sender):
            fk_id = getattr(instance, field.name)
            signal = f"{field.related_model_name}.{fk_id}.{field.relation_name}"
            signals.update((signal, f"{signal}.{action}"))

        if signals:
            self.extend([Signal({(signal, self.processing_component_id) for signal in signals})])

    def extend(self, commands: list[Command]):
        if commands:
            self._commands.extend(commands)
            self._optimize()

    def append(self, command: Command):
        self._commands.append(command)
        self._optimize()

    def pop(self) -> Command:
        return self._commands.pop(0)

    def __bool__(self):
        return bool(self._commands)

    def _optimize(self):
        self._commands.sort(key=self._priority)
        new_commands = []
        for i, command in enumerate(self._commands):
            match command:
                case (
                    Execute()
                    | Signal()
                    | Emit()
                    | SkipRender()
                    | Focus()
                    | Redirect()
                    | DispatchDOMEvent()
                    | Open()
                    | PushURL()
                    | ReplaceURL()
                ):
                    new_commands.append(command)

                case Destroy(component_id) as command:
                    # Register destroyed ids
                    self._destroyed_ids.add(component_id)
                    new_commands.append(command)

                case BuildAndRender(_, state, _) as command:
                    # Remove BuildAndRender of destroyed ids
                    if not (
                        (component_id := state.get("id")) and component_id in self._destroyed_ids
                    ):
                        new_commands.append(command)

                case Render(component) as command:
                    # Remove Render of destroyed ids
                    # Let the latest Render of the same component survive, kill the rest
                    if component.id not in self._destroyed_ids and not any(
                        isinstance(ahead_command, Render)
                        and ahead_command.component.id == component.id
                        and ahead_command.template is None
                        for ahead_command in self._commands[i + 1 :]
                    ):
                        new_commands.append(command)

        self._commands = new_commands

    @staticmethod
    def _priority(command: Command) -> tuple[int, str, int]:
        match command:
            case Execute():
                return 0, "", 0
            case Signal(_, timestamp):
                return 1, "", timestamp
            case Emit(_, timestamp):
                return 2, "", timestamp
            case Destroy():
                return 3, "", 0
            case SkipRender():
                return 4, "", 0
            case BuildAndRender(_, _, _, _, timestamp):
                return 5, "", timestamp
            case Render(component, template, _, _, _, timestamp):
                if template:
                    return 6, component.id, timestamp
                else:
                    return 7, component.id, timestamp
            case Focus() | Redirect() | DispatchDOMEvent() | Open() | ReplaceURL() | PushURL():
                return 8, "", 0
            case _ as unreachable:
                assert_never(unreachable)



from __future__ import annotations

import logging
import random
from collections import defaultdict
from collections.abc import AsyncIterable, Generator, Iterable
from dataclasses import dataclass
from dataclasses import field as Field
from typing import Any

from django.core.signing import Signer
from django.http import HttpRequest, QueryDict
from django.utils.html import format_html
from django.utils.safestring import SafeString, mark_safe
from pydantic import ValidationError
from uuid6 import uuid7

from djhtmx.global_events import HtmxUnhandledError
from djhtmx.tracing import tracing_span

from . import json
from .command_queue import CommandQueue
from .commands import PushURL, ReplaceURL, SendHtml
from .component import (
    LISTENERS,
    REGISTRY,
    BuildAndRender,
    Command,
    Destroy,
    DispatchDOMEvent,
    Emit,
    Execute,
    Focus,
    HtmxComponent,
    Open,
    Redirect,
    Render,
    Signal,
    SkipRender,
    _get_query_patchers,
)
from .introspection import filter_parameters
from .settings import (
    KEY_SIZE_ERROR_THRESHOLD,
    KEY_SIZE_SAMPLE_PROB,
    KEY_SIZE_WARN_THRESHOLD,
    LOGIN_URL,
    SESSION_TTL,
    conn,
)
from .utils import db, get_params

signer = Signer()

logger = logging.getLogger(__name__)


ProcessedCommand = (
    Destroy | Redirect | Open | Focus | DispatchDOMEvent | SendHtml | PushURL | ReplaceURL
)


class Repository:
    """An in-memory (cheap) mapping of component IDs to its states.

    When an HTMX request comes, all the state from all the components are
    placed in a registry.  This way we can instantiate components if/when
    needed.

    For instance, if a component is subscribed to an event and the event fires
    during the request, that component is rendered.

    """

    @staticmethod
    def new_session_id():
        return f"djhtmx:{uuid7().hex}"

    @classmethod
    def from_request(
        cls,
        request: HttpRequest,
    ) -> Repository:
        """Get or build the Repository from the request.

        If the request has already a Repository attached, return it without
        further processing.

        Otherwise, build the repository from the request's POST and attach it
        to the request.

        """
        from django.contrib.auth.models import AnonymousUser

        if (result := getattr(request, "htmx_repo", None)) is None:
            if (signed_session := request.META.get("HTTP_HX_SESSION")) and not bool(
                request.META.get("HTTP_HX_BOOSTED")
            ):
                session_id = signer.unsign(signed_session)
            else:
                session_id = cls.new_session_id()

            session = Session(session_id)

            result = cls(
                user=getattr(request, "user", AnonymousUser()),
                session=session,
                params=get_params(request),
            )
            request.htmx_repo = result  # type: ignore
        return result

    @classmethod
    def from_websocket(cls, user):
        return cls(
            user=user,
            session=Session(cls.new_session_id()),  # TODO: take the session from the websocket url
            params=get_params(None),
        )

    @staticmethod
    def load_states_by_id(states: list[str]) -> dict[str, dict[str, Any]]:
        return {
            state["id"]: state for state in [json.loads(signer.unsign(state)) for state in states]
        }

    @staticmethod
    def load_subscriptions(
        states_by_id: dict[str, dict[str, Any]], subscriptions: dict[str, str]
    ) -> dict[str, set[str]]:
        subscriptions_to_ids: dict[str, set[str]] = defaultdict(set)
        for component_id, component_subscriptions in subscriptions.items():
            # Register query string subscriptions
            component_name = states_by_id[component_id]["hx_name"]
            for patcher in _get_query_patchers(component_name):
                subscriptions_to_ids[patcher.signal_name].add(component_id)

            # Register other subscriptions
            for subscription in component_subscriptions.split(","):
                subscriptions_to_ids[subscription].add(component_id)
        return subscriptions_to_ids

    def __init__(
        self,
        user,
        session: Session,
        params: QueryDict,
    ):
        self.user = user
        self.session = session
        self.session_signed_id = signer.sign(session.id)
        self.params = params

    # Component life cycle & management

    def unregister_component(self, component_id: str):
        # delete component state
        self.session.unregister_component(component_id)

    async def adispatch_event(  # pragma: no cover
        self,
        component_id: str,
        event_handler: str,
        event_data: dict[str, Any],
    ) -> AsyncIterable[ProcessedCommand]:
        commands = CommandQueue([Execute(component_id, event_handler, event_data)])
        # Command loop
        try:
            while commands:
                processed_commands = self._run_command(commands)
                while command := await db(next)(processed_commands, None):
                    yield command
        except ValidationError as e:
            # This is here to detect validation errors derived from an invalid User
            # Meaning that the user type is not the right one so a login redirect has to happen
            if any(
                e
                for error in e.errors()
                if error["type"] == "is_instance_of" and error["loc"] == ("user",)
            ):
                yield Redirect(LOGIN_URL)
            else:
                raise

    def dispatch_event(
        self,
        component_id: str,
        event_handler: str,
        event_data: dict[str, Any],
    ) -> Iterable[ProcessedCommand]:
        commands = CommandQueue([Execute(component_id, event_handler, event_data)])

        # Command loop
        try:
            while commands:
                yield from self._run_command(commands)
        except ValidationError as e:
            # This is here to detect validation errors derived from an invalid User
            # Meaning that the user type is not the right one so a login redirect has to happen
            if any(
                e
                for error in e.errors()
                if error["type"] == "is_instance_of" and error["loc"] == ("user",)
            ):
                yield Redirect(LOGIN_URL)
            else:
                raise

    def _run_command(self, commands: CommandQueue) -> Generator[ProcessedCommand, None, None]:
        command = commands.pop()
        logger.debug("COMMAND: %s", command)
        commands_to_append: list[Command] = []
        match command:
            case Execute(component_id, event_handler, event_data):
                commands.processing_component_id = component_id
                match self.get_component_by_id(component_id):
                    case Destroy() as command:
                        yield command
                    case HtmxComponent() as component:
                        handler = getattr(component, event_handler)
                        handler_kwargs = filter_parameters(handler, event_data)
                        try:
                            emited_commands = handler(**handler_kwargs)
                        except Exception as error:
                            annotations = getattr(handler, "_htmx_annotations_", None)
                            logger.exception(
                                "HTMX unhandled exception in component %s",
                                component.__class__.__name__,
                            )
                            emited_commands = [
                                Emit(HtmxUnhandledError(error, handler_annotations=annotations))
                            ]
                        yield from self._process_emited_commands(
                            component, emited_commands, commands, during_execute=True
                        )

            case SkipRender(component):
                commands.processing_component_id = component.id
                self.session.store(component)

            case BuildAndRender(component_type, state, oob, parent_id):
                commands.processing_component_id = state.get("id", "")
                component = self.build(component_type.__name__, state)

                # Automatically track parent-child relationship if parent_id is specified
                child_id = component.id
                self.session.register_child(parent_id, child_id)

                commands_to_append.append(Render(component, oob=oob))

            case Render(component, template, oob, lazy, context):
                commands.processing_component_id = component.id
                html = self.render_html(
                    component, oob=oob, template=template, lazy=lazy, context=context
                )
                yield SendHtml(html, debug_trace=f"{component.hx_name}({component.id})")

            case Destroy(component_id) as command:
                commands.processing_component_id = component_id
                self.unregister_component(component_id)
                yield command

            case Emit(event):
                for component in self.get_components_by_names(*LISTENERS[type(event)]):
                    commands.processing_component_id = component.id
                    logger.debug("< AWAKED: %s id=%s", component.hx_name, component.id)
                    try:
                        emited_commands = component._handle_event(event)  # type: ignore
                    except Exception as error:
                        logger.exception(
                            "HTMX unhandled error in the event handler of %s",
                            component.__class__.__name__,
                        )
                        # Don't enter a spiral of death with HtmxUnhandledError
                        if not isinstance(event, HtmxUnhandledError):
                            emited_commands = [Emit(HtmxUnhandledError(error))]
                        else:
                            raise
                    yield from self._process_emited_commands(
                        component, emited_commands, commands, during_execute=False
                    )

            case Signal(signals):
                commands.processing_component_id = ""
                for component_or_destroy in self.get_components_subscribed_to(signals):
                    match component_or_destroy:
                        case Destroy() as command:
                            yield command
                        case component:
                            logger.debug("< AWAKED: %s id=%s", component.hx_name, component.id)
                            commands_to_append.append(Render(component))

            case (
                Open()
                | ReplaceURL()
                | PushURL()
                | Redirect()
                | Focus()
                | DispatchDOMEvent() as command
            ):
                commands.processing_component_id = ""
                yield command

        commands.extend(commands_to_append)
        self.session.flush()

    def _process_emited_commands(
        self,
        component: HtmxComponent,
        emmited_commands: Iterable[Command] | None,
        commands: CommandQueue,
        during_execute: bool,
    ) -> Iterable[ProcessedCommand]:
        component_was_rendered = False
        commands_to_add: list[Command] = []
        for command in emmited_commands or []:
            component_was_rendered = component_was_rendered or (
                isinstance(command, SkipRender | Render) and command.component.id == component.id
            )
            if (
                component_was_rendered
                and during_execute
                and isinstance(command, Render)
                and command.lazy is None
            ):
                # make partial updates not lazy during_execute
                command.lazy = False
            commands_to_add.append(command)

        if not component_was_rendered:
            commands_to_add.append(
                Render(component, lazy=False if during_execute else component.lazy)
            )

        if signals := self.update_params_from(component):
            yield ReplaceURL.from_params(self.params)
            commands_to_add.append(Signal({(signal, component.id) for signal in signals}))

        commands.extend(commands_to_add)
        self.session.store(component)

    def get_components_subscribed_to(
        self, signals: set[tuple[str, str]]
    ) -> Iterable[HtmxComponent | Destroy]:
        return (
            self.get_component_by_id(c_id)
            for c_id in sorted(self.session.get_component_ids_subscribed_to(signals))
        )

    def update_params_from(self, component: HtmxComponent) -> set[str]:
        """Updates self.params based on the state of the component

        Return the set of signals that should be triggered as the result of
        the update.

        """
        updated_params: set[str] = set()
        if patchers := _get_query_patchers(component.hx_name):
            for patcher in patchers:
                updated_params.update(
                    patcher.get_updates_for_params(
                        getattr(component, patcher.field_name, None),
                        self.params,
                    )
                )
        return updated_params

    def get_component_by_id(self, component_id: str) -> Destroy | HtmxComponent:
        """Return (possibly build) the component by its ID.

        If the component was already built, get it unchanged, otherwise build
        it from the request's payload and return it.

        If the `component_id` cannot be found, raise a KeyError.

        """
        if state := self.session.get_state(component_id):
            return self.build(state["hx_name"], state, retrieve_state=False)
        else:
            logger.error(
                "Component with id %s not found in session %s", component_id, self.session.id
            )
            return Destroy(component_id)

    def build(
        self,
        component_name: str,
        state: dict[str, Any],
        retrieve_state: bool = True,
        parent_id: str | None = None,
    ):
        """Build (or update) a component's state."""
        from django.contrib.auth.models import AnonymousUser

        with tracing_span("Repository.build", component_name=component_name):
            # Retrieve state from storage
            if retrieve_state and (component_id := state.get("id")):
                state = (self.session.get_state(component_id) or {}) | state

            # Patch it with whatever is the the GET params if needed
            for patcher in _get_query_patchers(component_name):
                state |= patcher.get_update_for_state(self.params)

            # Inject component name and user
            kwargs = state | {
                "hx_name": component_name,
                "user": None if isinstance(self.user, AnonymousUser) else self.user,
            }
            component = REGISTRY[component_name](**kwargs)

            # Automatically track parent-child relationship if parent_id is specified
            self.session.register_child(parent_id, component.id)

            return component

    def get_components_by_names(self, *names: str) -> Iterable[HtmxComponent]:
        # go over awaken components
        for name in names:
            for state in self.session.get_all_states():
                if state["hx_name"] == name:
                    yield self.build(name, {"id": state["id"]})

    def render_html(
        self,
        component: HtmxComponent,
        oob: str | None = None,
        template: str | None = None,
        lazy: bool | None = None,
        context: dict[str, Any] | None = None,
    ) -> SafeString:
        lazy = component.lazy if lazy is None else lazy
        with tracing_span(
            "Repository.render_html",
            component_name=component.hx_name,
            oob=str(oob),
            template=str(template),
            lazy=str(lazy),
        ):
            self.session.store(component)

            base_context = {
                "htmx_repo": self,
                "hx_oob": oob == "true",
                "this": component,
            }

            if lazy:
                template = template or component._template_name_lazy
                base_context |= {"hx_lazy": True} | component._get_lazy_context()
            else:
                base_context |= component._get_context()

            # If context is provided, it should override the component context
            if context is not None:
                # Keep base_context (htmx_repo, hx_oob, this) but let provided context override everything else
                final_context = {
                    "htmx_repo": self,
                    "hx_oob": oob == "true",
                    "this": component,
                } | context
            else:
                final_context = base_context

            html = mark_safe(component._get_template(template)(final_context).strip())

            # if performing some kind of append, the component has to be wrapped
            if oob and oob != "true":
                html = mark_safe(
                    "".join([
                        format_html('<div hx-swap-oob="{oob}">', oob=oob),
                        html,
                        "</div>",
                    ])
                )
            return html


@dataclass(slots=True)
class Session:
    id: str

    read: bool = False
    is_dirty: bool = False

    # dict[component_id -> state]
    states: dict[str, str] = Field(default_factory=dict)

    # dict[component_id -> set[signals]]
    subscriptions: defaultdict[str, set[str]] = Field(default_factory=lambda: defaultdict(set))

    # dict[parent_id -> set[child_ids]]
    children: defaultdict[str, set[str]] = Field(default_factory=lambda: defaultdict(set))

    # set[component_id]
    unregistered: set[str] = Field(default_factory=set)

    def store(self, component: HtmxComponent):
        state = component.model_dump_json()
        if self.states.get(component.id) != state:
            self.states[component.id] = state
            self.is_dirty = True

        subscriptions = component._get_all_subscriptions()
        if self.subscriptions[component.id] != subscriptions:
            self.subscriptions[component.id] = subscriptions
            self.is_dirty = True

    def unregister_component(self, component_id: str):
        # Recursively unregister all children first
        if child_ids := self.children.get(component_id):
            for child_id in child_ids.copy():  # Copy to avoid modification during iteration
                self.unregister_component(child_id)

        # Remove from parent's children list
        for child_ids in self.children.values():
            if component_id in child_ids:
                child_ids.remove(component_id)
                break

        # Remove this component's children mapping
        self.children.pop(component_id, None)

        # Remove component state and subscriptions
        self.states.pop(component_id, None)
        self.subscriptions.pop(component_id, None)
        self.unregistered.add(component_id)
        self.is_dirty = True

    def register_child(self, parent_id: str | None, child_id: str):
        """Register a parent-child relationship between components."""
        if parent_id and parent_id != child_id and child_id not in self.children[parent_id]:
            self.children[parent_id].add(child_id)
            self.is_dirty = True

    def get_state(self, component_id: str) -> dict[str, Any] | None:
        self._ensure_read()
        if state := self.states.get(component_id):
            return json.loads(state)

    def get_component_ids_subscribed_to(self, signals: set[tuple[str, str]]) -> Iterable[str]:
        self._ensure_read()
        for component_id, subscribed_to in self.subscriptions.items():
            # here we ignore signals emitted by the component it self
            if subscribed_to.intersection(signal for signal, cid in signals if cid != component_id):
                yield component_id

    def get_all_states(self) -> Iterable[dict[str, Any]]:
        self._ensure_read()
        return [json.loads(state) for state in self.states.values()]

    def _ensure_read(self):
        if not self.read:
            for component_id, state in conn.hgetall(f"{self.id}:states").items():  # type: ignore
                component_id = component_id.decode()
                if component_id == "__subs__":
                    # dict[component_id -> list[signals]]
                    for component_id, signals in json.loads(state).items():
                        self.subscriptions[component_id] = set(signals)
                elif component_id == "__children__":
                    # dict[parent_id -> list[child_ids]]
                    for parent_id, child_ids in json.loads(state).items():
                        self.children[parent_id] = set(child_ids)
                else:
                    self.states[component_id] = state.decode()
            self.read = True

    def flush(self, ttl: int = SESSION_TTL):
        if self.is_dirty:
            key = f"{self.id}:states"
            if self.unregistered:
                conn.hdel(key, *self.unregistered)
                self.unregistered.clear()
            if self.states:
                conn.hset(key, mapping=self.states)
            conn.hset(key, "__subs__", json.dumps(self.subscriptions))
            conn.hset(key, "__children__", json.dumps(self.children))
            conn.expire(key, ttl)
            # The command MEMORY USAGE is considered slow:
            # https://redis.io/docs/latest/commands/memory-usage/
            #
            # So we perform a trivial sampling with some prob to test the memory usage of the state.
            probe = random.random() <= KEY_SIZE_SAMPLE_PROB
            if probe and isinstance(usage := conn.memory_usage(key), int):
                if KEY_SIZE_ERROR_THRESHOLD and usage > KEY_SIZE_ERROR_THRESHOLD:
                    logger.error(
                        "HTMX session's size (%s) exceeded the size threshold %s",
                        usage,
                        KEY_SIZE_ERROR_THRESHOLD,
                    )
                elif KEY_SIZE_WARN_THRESHOLD and usage > KEY_SIZE_WARN_THRESHOLD:
                    logger.warning(
                        "HTMX session's size (%s) exceeded the size threshold %s",
                        usage,
                        KEY_SIZE_WARN_THRESHOLD,
                    )
            self.is_dirty = False



from datetime import timedelta

import redis
from django.conf import settings

VERSION = "2.0.4"
DEBUG = settings.DEBUG
CSRF_HEADER_NAME = settings.CSRF_HEADER_NAME[5:].replace("_", "-")
LOGIN_URL = settings.LOGIN_URL

SCRIPT_URLS = [
    f"htmx/{VERSION}/htmx{'' if DEBUG else '.min'}.js",
    "htmx/django.js",
]

DEFAULT_LAZY_TEMPLATE = getattr(settings, "DJHTMX_DEFAULT_LAZY_TEMPLATE", "htmx/lazy.html")
conn = redis.from_url(getattr(settings, "DJHTMX_REDIS_URL", "redis://localhost/0"))
SESSION_TTL = getattr(settings, "DJHTMX_SESSION_TTL", 3600)
if isinstance(SESSION_TTL, timedelta):
    SESSION_TTL = int(SESSION_TTL.total_seconds())


ENABLE_SENTRY_TRACING = getattr(settings, "DJHTMX_ENABLE_SENTRY_TRACING", True)
ENABLE_LOGFIRE_TRACING = getattr(settings, "DJHTMX_ENABLE_LOGFIRE_TRACING", False)


STRICT_EVENT_HANDLER_CONSISTENCY_CHECK = getattr(
    settings,
    "DJHTMX_STRICT_EVENT_HANDLER_CONSISTENCY_CHECK",
    False,
)

KEY_SIZE_ERROR_THRESHOLD = getattr(
    settings,
    "DJHTMX_KEY_SIZE_ERROR_THRESHOLD",
    0,
)
KEY_SIZE_WARN_THRESHOLD = getattr(
    settings,
    "DJHTMX_KEY_SIZE_WARN_THRESHOLD",
    50 * 1024,  # 50kb
)
KEY_SIZE_SAMPLE_PROB = getattr(
    settings,
    "DJHTMX_KEY_SIZE_SAMPLE_PROB",
    0.1,
)



from __future__ import annotations

import logging
import re
import time
from collections import defaultdict
from collections.abc import Callable
from dataclasses import dataclass
from dataclasses import field as dataclass_field
from functools import cache, cached_property, partial
from os.path import basename
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Literal,
    ParamSpec,
    TypeVar,
    cast,
    get_type_hints,
)

from django.db import models
from django.shortcuts import resolve_url
from django.template import Context, loader
from django.utils.safestring import SafeString, mark_safe
from pydantic import BaseModel, ConfigDict, Field, validate_call
from pydantic.fields import ModelPrivateAttr

from . import json, settings
from .commands import PushURL, ReplaceURL
from .exceptions import ComponentNotFound
from .introspection import (
    ModelConfig,
    Unset,
    annotate_model,
    get_event_handler_event_types,
    get_function_parameters,
)
from .query import Query, QueryPatcher
from .tracing import tracing_span
from .utils import generate_id

__all__ = ("ComponentNotFound", "HtmxComponent", "ModelConfig", "Query")


@dataclass(slots=True)
class Destroy:
    "Destroys the given component in the browser and in the caches."

    component_id: str
    command: Literal["destroy"] = "destroy"


@dataclass(slots=True)
class Redirect:
    "Executes a browser redirection to the given URL."

    url: str
    command: Literal["redirect"] = "redirect"

    @classmethod
    def to(cls, to: Callable[[], Any] | models.Model | str, *args, **kwargs):
        return cls(resolve_url(to, *args, **kwargs))


@dataclass(slots=True)
class Open:
    "Open a new window with the URL."

    url: str
    name: str = ""
    rel: str = "noopener noreferrer"
    target: str = "_blank"

    command: Literal["open-tab"] = "open-tab"

    @classmethod
    def to(cls, to: Callable[[], Any] | models.Model | str, *args, **kwargs):
        return cls(resolve_url(to, *args, **kwargs))


@dataclass(slots=True)
class Focus:
    "Executes a '.focus()' on the browser element that matches `selector`"

    selector: str
    command: Literal["focus"] = "focus"


@dataclass(slots=True)
class Execute:
    component_id: str
    event_handler: str
    event_data: dict[str, Any]


@dataclass(slots=True)
class DispatchDOMEvent:
    "Dispatches a DOM CustomEvent in the given target."

    target: str
    event: str
    detail: Any
    bubbles: bool = False
    cancelable: bool = False
    composed: bool = False
    command: Literal["dispatch_dom_event"] = "dispatch_dom_event"


@dataclass(slots=True)
class SkipRender:
    "Instruct the HTMX engine to avoid the render of the component."

    component: HtmxComponent


@dataclass(slots=True)
class BuildAndRender:
    component: type[HtmxComponent]
    state: dict[str, Any]
    oob: str = "true"
    parent_id: str | None = None
    timestamp: int = dataclass_field(default_factory=time.monotonic_ns)

    @classmethod
    def append(
        cls, target_: str, component_: type[HtmxComponent], parent_id: str | None = None, **state
    ):
        return cls(
            component=component_, state=state, oob=f"beforeend: {target_}", parent_id=parent_id
        )

    @classmethod
    def prepend(
        cls, target_: str, component_: type[HtmxComponent], parent_id: str | None = None, **state
    ):
        return cls(
            component=component_, state=state, oob=f"afterbegin: {target_}", parent_id=parent_id
        )

    @classmethod
    def after(
        cls, target_: str, component_: type[HtmxComponent], parent_id: str | None = None, **state
    ):
        return cls(
            component=component_, state=state, oob=f"afterend: {target_}", parent_id=parent_id
        )

    @classmethod
    def before(
        cls, target_: str, component_: type[HtmxComponent], parent_id: str | None = None, **state
    ):
        return cls(
            component=component_, state=state, oob=f"beforebegin: {target_}", parent_id=parent_id
        )

    @classmethod
    def update(cls, component: type[HtmxComponent], **state):
        return cls(component=component, state=state)


@dataclass(slots=True)
class Render:
    component: HtmxComponent
    template: str | None = None
    oob: str = "true"
    lazy: bool | None = None
    context: dict[str, Any] | None = None
    timestamp: int = dataclass_field(default_factory=time.monotonic_ns)


@dataclass(slots=True)
class Emit:
    "Emit a backend-only event."

    event: Any
    timestamp: int = dataclass_field(default_factory=time.monotonic_ns)


@dataclass(slots=True)
class Signal:
    "Emit a backend-only signal."

    names: set[tuple[str, str]]  # set[tuple[signal name, emitter component id]]
    timestamp: int = dataclass_field(default_factory=time.monotonic_ns)


Command = (
    Destroy
    | Redirect
    | Focus
    | DispatchDOMEvent
    | SkipRender
    | BuildAndRender
    | Render
    | Emit
    | Signal
    | Execute
    | Open
    | PushURL
    | ReplaceURL
)


RenderFunction = Callable[[Context | dict[str, Any] | None], SafeString]

PYDANTIC_MODEL_METHODS = {
    attr_name for attr_name in dir(BaseModel) if not attr_name.startswith("_")
}

REGISTRY: dict[str, type[HtmxComponent]] = {}
LISTENERS: dict[type, set[str]] = defaultdict(set)
FQN: dict[type[HtmxComponent], str] = {}


@cache
def _get_query_patchers(component_name: str) -> list[QueryPatcher]:
    return list(QueryPatcher.for_component(REGISTRY[component_name]))


@cache
def _get_querystring_subscriptions(component_name: str) -> frozenset[str]:
    return frozenset({
        patcher.signal_name
        for patcher in _get_query_patchers(component_name)
        if patcher.auto_subscribe
    })


A = TypeVar("A")
B = TypeVar("B")
P = ParamSpec("P")


def _compose(f: Callable[P, A], g: Callable[[A], B]) -> Callable[P, B]:
    def result(*args: P.args, **kwargs: P.kwargs):
        return g(f(*args, **kwargs))

    return result


RENDER_FUNC: dict[str, RenderFunction] = {}


def get_template(template: str) -> RenderFunction:  # pragma: no cover
    if settings.DEBUG:
        return cast(RenderFunction, _compose(loader.get_template(template).render, mark_safe))
    else:
        if (render := RENDER_FUNC.get(template)) is None:
            render = cast(RenderFunction, _compose(loader.get_template(template).render, mark_safe))
            RENDER_FUNC[template] = render
        return render


class HtmxComponent(BaseModel):
    _template_name: str = ...  # type: ignore
    _template_name_lazy: str = settings.DEFAULT_LAZY_TEMPLATE

    # tracks which attributes are properties, to expose them in a lazy way to the _get_context
    # during rendering
    _properties: set[str] = ...  # type: ignore

    # tracks what are the names of the event handlers of the class
    _event_handler_params: dict[str, frozenset[str]] = ...  # type: ignore

    # fields to exclude from component state during serialization
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )

    def __init_subclass__(cls, public=None):
        FQN[cls] = f"{cls.__module__}.{cls.__name__}"

        component_name = cls.__name__

        if public is None:
            # Detect concrete versions of generic classes, they are non public
            if "[" in component_name and "]" in component_name:
                public = False
            elif _ABSTRACT_BASE_REGEX.match(component_name):
                logger.info(
                    "HTMX Component: <%s> Automatically detected as non public",
                    FQN[cls],
                )
                public = False
            else:
                public = True

        if public:
            REGISTRY[component_name] = cls

            # Warn of components that do not have event handlers and are public
            if (
                not any(cls.__own_event_handlers(get_parent_ones=True))
                and not hasattr(cls, "_handle_event")
                and not hasattr(cls, "subscriptions")
            ):
                logger.warning(
                    "HTMX Component <%s> has no event handlers, probably should not exist and be just a template",
                    FQN[cls],
                )

        assert isinstance(cls._template_name, ModelPrivateAttr)
        if isinstance(cls._template_name.default, str) and (
            basename(cls._template_name.default)
            not in (f"{klass.__name__}.html" for klass in cls.__mro__)
        ):
            logger.warning(
                "HTMX Component <%s> template name does not match the component name",
                FQN[cls],
            )

        # We use 'get_type_hints' to resolve the forward refs if needed, but
        # we only need to rewrite the actual annotations of the current class,
        # that's why we iter over the '__annotations__' names.
        hints = get_type_hints(cls, include_extras=True)
        for name in list(cls.__annotations__):
            if not name.startswith("_"):
                annotation = hints[name]
                cls.__annotations__[name] = annotate_model(annotation)

        cls._event_handler_params = {
            name: get_function_parameters(event_handler)
            for name, event_handler in cls.__own_event_handlers(get_parent_ones=True)
        }

        for name, params in cls._event_handler_params.items():
            if params and not hasattr((attr := getattr(cls, name)), "raw_function"):
                setattr(
                    cls,
                    name,
                    validate_call(config={"arbitrary_types_allowed": True})(attr),
                )

        cls.__check_consistent_event_handler(strict=settings.STRICT_EVENT_HANDLER_CONSISTENCY_CHECK)
        if public:
            if handle_event := getattr(cls, "_handle_event", None):
                for event_type in get_event_handler_event_types(handle_event):
                    LISTENERS[event_type].add(component_name)

            cls._properties = {
                attr
                for attr in dir(cls)
                if not attr.startswith("_")
                if attr not in PYDANTIC_MODEL_METHODS
                if isinstance(getattr(cls, attr), property | cached_property)
            }

        return super().__init_subclass__()

    @classmethod
    def __own_event_handlers(cls, get_parent_ones=False):
        attr_names = dir(cls) if get_parent_ones else vars(cls)
        for attr_name in attr_names:
            if (
                not attr_name.startswith("_")
                and attr_name not in PYDANTIC_MODEL_METHODS
                and attr_name.islower()
                and callable(attr := getattr(cls, attr_name))
            ):
                yield attr_name, attr

    @classmethod
    def __check_consistent_event_handler(cls, *, strict: bool = False):
        """Check that '_handle_event' is consistent.

        If the class inherits from one that super-class, and it gets
        `_handle_event` from several of those branches, it must override it to
        resolve the ambiguity.

        Raise an error if there is no self-defined method.

        """
        parents = {
            method
            for base in cls.__bases__
            if (method := getattr(base, "_handle_event", None)) is not None
        }
        if len(parents) > 1:
            resolved = cls._handle_event  # type: ignore
            if resolved in parents:
                bases = ", ".join(
                    base.__name__
                    for base in cls.__bases__
                    if (method := getattr(base, "_handle_event", None)) is not None
                )
                if strict:
                    raise TypeError(
                        f"Component {cls.__name__} doesn't override "
                        f"_handle_event to reconcile the base classes ({bases})."
                    )
                else:
                    logger.error(
                        "Component %s doesn't override _handle_event to reconcile the base classes (%s)",
                        cls.__name__,
                        bases,
                    )

    # State
    id: Annotated[str, Field(default_factory=generate_id)]

    user: Annotated[Any | None, Field(exclude=True)]  # type: ignore
    if TYPE_CHECKING:
        from django.contrib.auth.models import AbstractBaseUser

        user: Annotated[AbstractBaseUser | None, Field(exclude=True)]

    hx_name: str
    lazy: bool = False

    def __repr__(self) -> str:
        return f"{self.hx_name}(\n{self.model_dump_json(indent=2, exclude={'hx_name'})})\n"

    @property
    def subscriptions(self) -> set[str]:
        return set()

    def render(self): ...

    def _get_all_subscriptions(self) -> set[str]:
        return self.subscriptions | _get_querystring_subscriptions(self.hx_name)

    def _get_template(self, template: str | None = None) -> Callable[..., SafeString]:
        return get_template(template or self._template_name)

    def _get_lazy_context(self):
        return {}

    def _get_context(self):
        # This render-local cache, supports lazy properties but avoids the same property to be
        # computed more than once.  It doesn't survive several renders which is good, because it
        # doesn't require invalidation.
        def get_property(cache, attr):
            result = cache.get(attr, Unset)
            if result is Unset:
                result = getattr(self, attr)
                cache[attr] = result
            return result

        with tracing_span(f"{FQN[type(self)]}._get_context"):
            render_cache = {}
            return {
                attr: (
                    partial(get_property, render_cache, attr)  # do lazy evaluation of properties
                    if attr in self._properties
                    else getattr(self, attr)
                )
                for attr in dir(self)
                if not attr.startswith("_") and attr not in PYDANTIC_MODEL_METHODS
            }


@dataclass(slots=True)
class Triggers:
    """HTMX triggers.

    Allow to trigger events on the client from the server.  See
    https://htmx.org/attributes/hx-trigger/

    """

    _trigger: dict[str, list[Any]] = dataclass_field(default_factory=lambda: defaultdict(list))
    _after_swap: dict[str, list[Any]] = dataclass_field(default_factory=lambda: defaultdict(list))
    _after_settle: dict[str, list[Any]] = dataclass_field(default_factory=lambda: defaultdict(list))

    def add(self, name, what: Any):
        self._trigger[name].append(what)

    def after_swap(self, name, what: Any):
        self._after_swap[name].append(what)

    def after_settle(self, name, what: Any):
        self._after_settle[name].append(what)

    @property
    def headers(self):
        headers = [
            ("HX-Trigger", self._trigger),
            ("HX-Trigger-After-Swap", self._after_swap),
            ("HX-Trigger-After-Settle", self._after_settle),
        ]
        return {header: json.dumps(value) for header, value in headers if value}


F = TypeVar("F")


def annotated_handler(**annotations) -> Callable[[F], F]:
    """Annotate the HTMX handler with customized values.

    Some of these annotations are HtmxUnhandledError use the annotations so that the application can
    have more detailed error recovery handlers.

    """

    def decorator(fn):
        if not hasattr(fn, "_htmx_annotations_"):
            fn._htmx_annotations_ = htmx_annotations = {}
        else:
            htmx_annotations = fn._htmx_annotations_
        htmx_annotations.update(annotations)
        return fn

    return decorator


logger = logging.getLogger(__name__)


_ABSTRACT_BASE_REGEX = re.compile(r"^(_)?(Base|Abstract)[A-Z0-9_]")



import logging
from typing import Any, Literal, assert_never

from channels.generic.websocket import AsyncJsonWebsocketConsumer
from pydantic import BaseModel, TypeAdapter

from . import json
from .commands import PushURL, ReplaceURL, SendHtml
from .component import Command, Destroy, DispatchDOMEvent, Focus, Open, Redirect
from .introspection import parse_request_data
from .repo import Repository
from .utils import get_params


class ComponentsRemoved(BaseModel):
    type: Literal["removed"]
    component_ids: list[str]


class ComponentsAdded(BaseModel):
    type: Literal["added"]
    states: list[str]
    subscriptions: dict[str, str]


Event = ComponentsRemoved | ComponentsAdded
EventAdapter = TypeAdapter(Event)


class Consumer(AsyncJsonWebsocketConsumer):
    async def connect(self):
        await self.accept()
        self.repo = Repository.from_websocket(self.scope["user"])

    async def disconnect(self, code):
        await super().disconnect(code)

    async def receive_json(self, event_data: dict[str, Any]):
        if headers := event_data.pop("HEADERS", None):
            url = headers["HX-Current-URL"]
            component_id = headers["HX-Component-Id"]
            event_handler = headers["HX-Component-Handler"]
            params = get_params(url)
            logger.debug(">>>> Call: %s %s", component_id, event_handler)
            self.repo.params.clear()
            self.repo.params.update(params)  # type: ignore
            # Command dispatching
            async for command in self.repo.adispatch_event(
                component_id, event_handler, parse_request_data(event_data)
            ):
                match command:
                    case SendHtml(html, debug_trace):
                        logger.debug(
                            "< Command: %s", f"SendHtml[{debug_trace}](... {len(html)} ...)"
                        )
                        await self.send(html)
                    case (
                        Destroy()
                        | Redirect()
                        | Focus()
                        | DispatchDOMEvent()
                        | PushURL()
                        | Open()
                        | ReplaceURL()
                    ):
                        logger.debug("< Command: %s", command)
                        await self.send_json(command)
                    case _ as unreachable:
                        assert_never(unreachable)

    async def send_commands(self, commands: list[Command]):
        for command in commands:
            await self.send_json(command)

    @classmethod
    async def decode_json(cls, text_data) -> dict[str, Any]:
        return json.loads(text_data)

    @classmethod
    async def encode_json(cls, content) -> str:
        return json.dumps(content)


logger = logging.getLogger(__name__)
